/* $Workfile: T2p5x16Hdlc.c $       $Revision: 30 $      $Modtime: 5/18/01 10:26a $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T2P5X16HDLC.C

    A example console application demonstrating the use of HDLC pipes connected
    over the H.100 bus between two OTX PCI boards (one THOR-2-PCI board (or one 
    THOR-8-PCI) and one Vidar-5x16-PCI board).

    The Thor-2-PCI board hosts two E1/T1 transceivers. The Thor-8-PCI board hosts 
    eight E1/T1 tranceivers.

    The E1/T1 transceivers are configured for E1 mode.

    The DSPs on the Vidar-5x16-PCI board hosts a number of 64kpbs HDLC Senders and 
    Receivers implemented in the OtxHdlc.out DSP program. Each of the 16 DSPs on 
    the Vidar-5x16-PCI is configured for 4 HDLC Senders and 4 HDLC Receivers 
    (a total of 64 HDLC Senders and 64 HDLC Receivers). 

    Pipes are connected to and from the E1T1 devices on the Thor-2-PCI (or Thor-8-PCI)
    boards to the DSPs on the Vidar-5x16-PCI board (via the H.100 bus).

    Thor-2-PCI: The HDLC Sender/Receivers are connected to E1 TS0-TS31 of each Li 
    (Li0 and Li1). The HDLC Sender/Receiver connected to E1 TS0 is never enabled since
    E1 TS0 is the E1 synchronization timeslot.
    
    Thor-8-PCI: The HDLC Sender/Receivers are connected to E1 TS0-TS7 of each Li 
    (Li0-Li7). The HDLC Sender/Receiver connected to E1 TS0 is never enabled since
    E1 TS0 is the E1 synchronization timeslot.
    
    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call
#define DEMO_DSP_MAX 16
#define DEMO_LOGIC_DEVS_MAX 64
#define DEMO_DEFAULT_BOARD_NO 0

#define DEMO_SPM1_HDLC_SENDER 1        // Comment this line to remove the HDLC Sender Logical devices from this application

#define DEMO_DEFAULT_LOG_FILE "T2p5x16Hdlc.log"

// This example program uses a fixed connection method to the H.100 bus by opening the 
// Time-Space Switch (TSS) physical device. 
// The TSS device has 16 local highways, and 32 H.100 highways. To make a connection
// between a device (e.g. DSP) on a local highway and one H.100 highway, a pipe is first
// created. Then one end is connected the the DSP device on the local highway. The other
// end connects to TSS device on one of the H.100 highways.

// Note that the local highways are bi-directional (i.e. you can have a transmit 
// and a receive pipe on one timeslot. The H.100 highways are uni-directional 
// (i.e. you need one timeslot for transmit and another timeslot for receive).
// 
// The OTX driver also offers another simplified way of connecting between
// two OTX boards. The pipe can be directly connect from a physical device on one board
// to physical device on another board (without having to open and connect to the TSS device).
// However, this demo program does not use this feature.
#define DEMO_LI0_HIGHWAY      0
#define DEMO_LI1_HIGHWAY      1
#define DEMO_DSP0_HIGHWAY     0
#define DEMO_DSP1_HIGHWAY     1
#define DEMO_CODEC_HIGHWAY   14    
#define DEMO_H100_HWY_0      16
#define DEMO_H100_HWY_1      17
#define DEMO_H100_HWY_2      18
// ....
#define DEMO_H100_HWY_31     47



typedef struct DemoEnvT {
  char           *logFileName;
  OtxT1E1LiModeE  eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL        bSs7FisuFiltering;
  int             boardNo;
  int             coLi;
  OTX_INT32       nBoardType;
} DemoEnvT;


/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoInitalizeGlobalData(void);
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoOpenPhysicalDevices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoPrintDeviceState(OTX_HANDLE hHandle);
static void DemoLogf(char *formatStr, ...);
static void DemoHelpPanel(void);
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoPrintTimeStamp(OTX_DATETIME nTimeStamp);


/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;
OTX_INT32   g_nMsgLen = 10;
OTX_BOOL    g_bHdlcVerbose = OTX_TRUE;
OTX_UINT32  g_coHdlcRxFrames[DEMO_LOGIC_DEVS_MAX];


struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hTss;
  OTX_HANDLE hQuadE1T1[2]; // Used for Thor-8-PCI (not for Thor-2-PCI)
  OTX_HANDLE hLi[8];
} g_hPhysDevsThor;


struct {
  OTX_HANDLE hVidarController0;
  OTX_HANDLE hTss;
  OTX_HANDLE hDsp[16];
} g_hPhysDevsVidar;


struct {
  OTX_HANDLE hHandle;

  // Pipes
  OTX_HANDLE hH100LiPipe;
  OTX_HANDLE hDspH100Pipe;

  OTX_INT32 nDsp;
  OTX_INT32 nLi;
  OTX_INT32 nThorHwy;
  OTX_INT32 nThorTs;
  OTX_INT32 nVidarHwy;
  OTX_INT32 nVidarTs;
  OTX_INT32 nH100Hwy;
  OTX_INT32 nH100Ts;

} g_sqSenders[DEMO_LOGIC_DEVS_MAX];

struct {
  OTX_HANDLE hHandle;

  // Pipes
  OTX_HANDLE hLiH100Pipe;
  OTX_HANDLE hH100DspPipe;

  OTX_INT32 nDsp;
  OTX_INT32 nLi;
  OTX_INT32 nThorHwy;
  OTX_INT32 nThorTs;
  OTX_INT32 nVidarHwy;
  OTX_INT32 nVidarTs;
  OTX_INT32 nH100Hwy;
  OTX_INT32 nH100Ts;

} g_sqReceivers[DEMO_LOGIC_DEVS_MAX];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func


*******************************************************************************/
void DemoInitalizeGlobalData(void)
{
  OTX_INT32  i;
  OTX_INT32 nChannelsPerLi = DEMO_LOGIC_DEVS_MAX / g_demoEnv.coLi;
  OTX_INT32 nChannelsPerDsp = DEMO_LOGIC_DEVS_MAX / DEMO_DSP_MAX;

  memset(g_sqReceivers,    0, sizeof(g_sqReceivers));
  memset(g_sqSenders, 0, sizeof(g_sqSenders));

  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {

    g_sqSenders[i].nDsp      = i / nChannelsPerDsp;
    g_sqSenders[i].nLi       = i / nChannelsPerLi;
    g_sqSenders[i].nThorHwy  = g_sqSenders[i].nLi;
    g_sqSenders[i].nThorTs   = i % nChannelsPerLi;
    g_sqSenders[i].nVidarHwy = (g_sqSenders[i].nDsp < 8) ? g_sqSenders[i].nDsp /* 0-7 */ : ((g_sqSenders[i].nDsp - 8) / 4)  + 8 /* 8, 9 */;
    g_sqSenders[i].nVidarTs  = (g_sqSenders[i].nDsp < 8) ? i % nChannelsPerDsp /* 0-3 */: (i % nChannelsPerDsp) + (8*(g_sqSenders[i].nDsp - 8)%32) /* 0-3, 8-11, 16-19, 24-27 */;
    g_sqSenders[i].nH100Hwy  = DEMO_H100_HWY_0 + (i / 32);
    g_sqSenders[i].nH100Ts   = i % 32;

    g_sqReceivers[i].nDsp      = g_sqSenders[i].nDsp;
    g_sqReceivers[i].nLi       = g_sqSenders[i].nLi;
    g_sqReceivers[i].nThorHwy  = g_sqSenders[i].nThorHwy;
    g_sqReceivers[i].nThorTs   = g_sqSenders[i].nThorTs;
    g_sqReceivers[i].nVidarHwy = g_sqSenders[i].nVidarHwy;
    g_sqReceivers[i].nVidarTs  = g_sqSenders[i].nVidarTs;
    g_sqReceivers[i].nH100Hwy  = DEMO_H100_HWY_0 + (i / 32) + (DEMO_LOGIC_DEVS_MAX / 32);
    g_sqReceivers[i].nH100Ts   = g_sqSenders[i].nH100Ts;

    
  }

#if 1
  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
    DemoLogf("TX[%.2d] Dsp=%.2d Li=%.2d ThHwy=%.2d ThTs=%.2d VdHwy=%.2d VdTs=%.2d HHwy=%.2d HTs=%.2d\n", 
      i,
      g_sqSenders[i].nDsp,      
      g_sqSenders[i].nLi,       
      g_sqSenders[i].nThorHwy, 
      g_sqSenders[i].nThorTs,     
      g_sqSenders[i].nVidarHwy, 
      g_sqSenders[i].nVidarTs,  
      g_sqSenders[i].nH100Hwy,  
      g_sqSenders[i].nH100Ts
      );
  }
  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
    DemoLogf("RX[%.2d] Dsp=%.2d Li=%.2d ThHwy=%.2d ThTs=%.2d VdHwy=%.2d VdTs=%.2d HHwy=%.2d HTs=%.2d\n", 
      i,
      g_sqReceivers[i].nDsp,      
      g_sqReceivers[i].nLi,       
      g_sqReceivers[i].nThorHwy, 
      g_sqReceivers[i].nThorTs,     
      g_sqReceivers[i].nVidarHwy, 
      g_sqReceivers[i].nVidarTs,  
      g_sqReceivers[i].nH100Hwy,  
      g_sqReceivers[i].nH100Ts
      );
  }
#endif

}

/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent);
  if (result == OTX_E_NOTIMPL) {
    g_hNotificationEvent = NULL;    // D16V86 apps cannot use event objects
    result = OTX_S_OK;
  }

  DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor-2-PCI board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;
  OTX_INT32  nQuadE1T1; // Only used for Thor-8-PCI

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening Thor-2-PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI:
    DemoLogf("Opening Thor-8-PCI board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, (OtxPhysicalDeviceTypeE) g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevsThor.hThorController0); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);

  // Get a handle to the Time-Space Switch (TSS) on the Thor-2-PCI board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor.hThorController0, OTX_DEVICE_TSS, 0, 0, g_hEventQueue, &g_hPhysDevsThor.hTss); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);


  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    // On the Thor-2-PCI board the E1T1 devices are hosted by the board itself. Get handles to T1/E1 line interface devices #0 and #1
    for (i = 0; i < g_demoEnv.coLi; i++) {
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d\n", i);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor.hThorController0, OTX_DEVICE_LI_T1E1, i, i, g_hEventQueue, &g_hPhysDevsThor.hLi[i]); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);
    }
    break;
  case OTX_DEVICE_THOR_8_PCI:
    // On the Thor-2-PCI board the E1T1 devices are hosted by a quad E1T1 device. Open those two devices.
    for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
      DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevsThor.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);
    }

    // Get handles to T1/E1 line interface devices 
    for (i = 0; i < g_demoEnv.coLi; i++) {
      nQuadE1T1 = i/4;
      DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
      result = OtxDrvOpenPhysicalDevice(g_hPhysDevsThor.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevsThor.hLi[i]); DemoCheckRc(g_hPhysDevsThor.hQuadE1T1[nQuadE1T1], result);
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Connect to the Physical Vidar-5x16 Board Controller
  DemoLogf("Open OTX_DEVICE_VIDAR_5x_PCI (Vidar-5x16-PCI board)\n");
  result = OtxDrvOpenPhysicalDevice(0, OTX_DEVICE_VIDAR_5x_PCI, g_demoEnv.boardNo, g_demoEnv.boardNo + 10, g_hEventQueue, &g_hPhysDevsVidar.hVidarController0); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvOpenPhysicalDevice(g_hPhysDevsVidar.hVidarController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevsVidar.hDsp[i]); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);
  }


  // Get a handle to the Time-Space Switch (TSS) on the Vidar-5x16-PCI board
  DemoLogf("Open OTX_DEVICE_TSS\n");
  result = OtxDrvOpenPhysicalDevice(g_hPhysDevsVidar.hVidarController0, OTX_DEVICE_TSS, 0, 1, g_hEventQueue, &g_hPhysDevsVidar.hTss); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);


  // Load DSP Program
  for (i = 0; i < DEMO_DSP_MAX; i++) {
    DemoLogf("Loading OTX_DSP_SPM_OTXHDLC to DSP#%d \n", i);
    result = OtxDspRunSpm(g_hPhysDevsVidar.hDsp[i], OTX_DSP_SPM_OTXHDLC); DemoCheckRc(g_hPhysDevsVidar.hDsp[i], result);
  }

}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevsThor.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevsThor.hLi[i], result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevsThor.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevsThor.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevsThor.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevsThor.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevsThor.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevsThor.hLi[i], result);

#if 1 
    {    
      struct {
        OtxE1ConfigOptionsS e1CfgData;
        OtxT1ConfigOptionsS t1CfgData;
      } g_T1E1CfgData2;

      result = OtxE1GetConfiguration(g_hPhysDevsThor.hLi[i], &g_T1E1CfgData2.e1CfgData); DemoCheckRc(g_hPhysDevsThor.hLi[i], result);
    }
#endif

  }

}

/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevsThor.hThorController0,   OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);
  result = OtxDrvEnable(g_hPhysDevsVidar.hVidarController0, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);

  
  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevsThor.hLi[i],           OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevsThor.hLi[i], result);
  }
  

  for (i = 0; i < DEMO_DSP_MAX; i++) {
    result = OtxDrvEnable(g_hPhysDevsVidar.hDsp[i],         OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevsVidar.hDsp[i], result);
  }
}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
#ifdef DEMO_SPM1_HDLC_SENDER
    result = OtxDrvCreateLogicalDevice(g_hPhysDevsVidar.hDsp[g_sqSenders[i].nDsp], OTX_LDEVICE_HDLC_SENDER,   g_hEventQueue, i, &g_sqSenders[i].hHandle); DemoCheckRc(g_hPhysDevsVidar.hDsp[g_sqSenders[i].nDsp], result);
#endif
    result = OtxDrvCreateLogicalDevice(g_hPhysDevsVidar.hDsp[g_sqReceivers[i].nDsp], OTX_LDEVICE_HDLC_RECEIVER, g_hEventQueue, i, &g_sqReceivers[i].hHandle); DemoCheckRc(g_hPhysDevsVidar.hDsp[g_sqReceivers[i].nDsp], result);
  }


}



/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    by its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  //OTX_INT32 ts;
  //OTX_INT32 h100Hwy;
  //OTX_INT32 localHwy;
  //OTX_INT32 nDsp;
	//OTX_INT32 nLi;


#if 1
	// Vidar connections
  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
    
    result = OtxDrvCreatePipe(64, &g_sqSenders[i].hDspH100Pipe); DemoCheckRc(g_sqSenders[i].hDspH100Pipe, result);
    result = OtxDrvCreatePipe(64, &g_sqReceivers[i].hH100DspPipe); DemoCheckRc(g_sqReceivers[i].hH100DspPipe, result);
    
    // The local highways are bi-directional. I.e. we use the same TS for both TX and RX directions
    // The H.100 highways are uni-directional. I.e. we must have a separate TS for RX and TX directions
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqSenders[i].nVidarTs] = 0xFF;
    result = OtxDrvConnectPipe(g_sqSenders[i].hDspH100Pipe, OTX_PIPE_INPUT,  g_hPhysDevsVidar.hDsp[g_sqSenders[i].nDsp],  g_sqSenders[i].nVidarHwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqSenders[i].hDspH100Pipe, result);
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqSenders[i].nH100Ts] = 0xFF;
    result = OtxDrvConnectPipe(g_sqSenders[i].hDspH100Pipe, OTX_PIPE_OUTPUT, g_hPhysDevsVidar.hTss, g_sqSenders[i].nH100Hwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqSenders[i].hDspH100Pipe, result);
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqReceivers[i].nH100Ts] = 0xFF;
    result = OtxDrvConnectPipe(g_sqReceivers[i].hH100DspPipe, OTX_PIPE_INPUT,  g_hPhysDevsVidar.hTss, g_sqReceivers[i].nH100Hwy,  sqTimeSlotMask, 32); DemoCheckRc(g_sqReceivers[i].hH100DspPipe, result);
    
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqReceivers[i].nVidarTs] = 0xFF;
    result = OtxDrvConnectPipe(g_sqReceivers[i].hH100DspPipe, OTX_PIPE_OUTPUT, g_hPhysDevsVidar.hDsp[g_sqReceivers[i].nDsp],  g_sqReceivers[i].nVidarHwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqReceivers[i].hH100DspPipe, result);
    
    //DemoLogf("DSP%d OUTPUT (HWY#%d TS#%d) connected\n", nDsp, localHwy, ts); 
    
    // Create the following Connections in receive direction:
    //  TSS  ------> HDLC Receiver
    //DemoLogf("TSS  ------> HDLC Receiver\n"); 
    result = OtxDrvConnectLogicalDevice(g_sqReceivers[i].hH100DspPipe, -1, g_sqReceivers[i].hHandle,  -1); DemoCheckRc(g_sqReceivers[i].hH100DspPipe, result);
    
    // Make the following connections in transmit direction:
    // HDLC Sender   -----> TSS
#ifdef DEMO_SPM1_HDLC_SENDER
    //DemoLogf("HDLC Sender   -----> TSS\n"); 
    result = OtxDrvConnectLogicalDevice(g_sqSenders[i].hHandle, -1, g_sqSenders[i].hDspH100Pipe, -1); DemoCheckRc(g_sqSenders[i].hHandle, result);
#endif
    
  } 
#endif

  // Thor connections
  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
    result = OtxDrvCreatePipe(64, &g_sqSenders[i].hH100LiPipe); DemoCheckRc(g_sqSenders[i].hH100LiPipe, result);
    result = OtxDrvCreatePipe(64, &g_sqReceivers[i].hLiH100Pipe); DemoCheckRc(g_sqReceivers[i].hLiH100Pipe, result);
    
    // The local highways are bi-directional. I.e. we use the same TS for both TX and RX directions
    // The H.100 highways are uni-directional. I.e. we must have a separate TS for RX and TX directions
    
    // TX
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqSenders[i].nH100Ts] = 0xFF;
    result = OtxDrvConnectPipe(g_sqSenders[i].hH100LiPipe, OTX_PIPE_INPUT,  g_hPhysDevsThor.hTss, g_sqSenders[i].nH100Hwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqSenders[i].hH100LiPipe, result);
    // TX
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqSenders[i].nThorTs] = 0xFF;
    result = OtxDrvConnectPipe(g_sqSenders[i].hH100LiPipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor.hTss,  g_sqSenders[i].nThorHwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqSenders[i].hH100LiPipe, result);
    // RX
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqReceivers[i].nThorTs] = 0xFF;
    result = OtxDrvConnectPipe(g_sqReceivers[i].hLiH100Pipe, OTX_PIPE_INPUT,  g_hPhysDevsThor.hTss,  g_sqReceivers[i].nThorHwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqReceivers[i].hLiH100Pipe, result);
    // RX
    memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
    sqTimeSlotMask[g_sqReceivers[i].nH100Ts] = 0xFF;
    result = OtxDrvConnectPipe(g_sqReceivers[i].hLiH100Pipe, OTX_PIPE_OUTPUT, g_hPhysDevsThor.hTss, g_sqReceivers[i].nH100Hwy, sqTimeSlotMask, 32); DemoCheckRc(g_sqReceivers[i].hLiH100Pipe, result);
    
  } 
  

      
}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
    
#ifdef DEMO_SPM1_HDLC_SENDER
    // Initialize the Hdlc Sender 
    result = OtxHdlcSenderInit(
      g_sqSenders[i].hHandle, // Handle to logical HDLC Sender Device
      OTX_HDLC_FFT_FLAGS/*OTX_HDLC_FFT_ALL_ONES*/,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
      OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
      ); DemoCheckRc(g_sqSenders[i].hHandle, result);          
#endif
    
    // Initialize the Hdlc Receiver 
    result = OtxHdlcReceiverInit(
      g_sqReceivers[i].hHandle, // Handle to logical HDLC Receiver Device
      OTX_HDLC_FFT_FLAGS/*OTX_HDLC_FFT_ALL_ONES*/,        // Fill character(s) to be send between HDLC frames
      2,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
      OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
      OTX_FALSE,                 // If OTX_TRUE then all bits are inverted when transmitted and before received
      OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
      OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
      OTX_TRUE                   // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
      ); DemoCheckRc(g_sqReceivers[i].hHandle, result);          
    
    if (g_demoEnv.bSs7FisuFiltering) {
      result = OtxHdlcSS7SetFilter(g_sqReceivers[i].hHandle, OTX_HDLC_SS7_FILTER_FISU); DemoCheckRc(g_sqReceivers[i].hHandle, result);
    }
    
  }

}


/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
    // Enable Logical Devices
    if (g_sqReceivers[i].nThorTs == 0) {
      DemoLogf("Skipping enabling HDLC Receiver[%d] (because nThorTs=%d)\n", i, g_sqReceivers[i].nThorTs); 
      continue;
    }
    result = OtxDrvEnable(g_sqReceivers[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqReceivers[i].hHandle, result);
#ifdef DEMO_SPM1_HDLC_SENDER
    result = OtxDrvEnable(g_sqSenders[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqSenders[i].hHandle, result);
#endif
  }

}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eRxLineCode                = OTX_E1_LC_HDB3;
  pE1CfgData->eTxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eRxFrameFormat             = OTX_E1_FF_CRC4_MULTIFRAME;
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eRxLineCode                = OTX_T1_LC_B8ZS;
  pT1CfgData->eFrameFormat               = OTX_T1_FF_ESF;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000;
}

/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct tm *ptCurrentDayAndTime;
  time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
		nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif
  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
  ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
}

static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  short c;
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_SUCCESS(otxRc)) {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d): %s \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
      } else {
        DemoLogf("ERROR(%#lX) Dev(%X,%d,%d) Cannot translate this result code \n", result, hDevice, (int) eDeviceType, (int)nDeviceNumber);
      }
    }


    DemoLogf("Do you want to exit (Y/N) ");
    c = getch();
    if (c == 'y' || c == 'Y') {
      // Clean up
      OtxDrvReleaseLib();
      OsaConsoleExit();
			exit(1);
    }

  }
}


/*******************************************************************************

  @func
    Print the state of a device

  @comm

*******************************************************************************/
void DemoPrintDeviceState(
  OTX_HANDLE hHandle
)
{
  OTX_RESULT result;
  OtxDeviceStateE  eDevState;
  char *szDevState;

  result = OtxDrvGetState(hHandle, &eDevState); DemoCheckRc(hHandle, result);
  switch (eDevState) {
  case OTX_DEVICE_STATE_UNKNOWN:         
    szDevState = "OTX_DEVICE_STATE_UNKNOWN";
    break;
  case OTX_DEVICE_WORKING:
    szDevState = "OTX_DEVICE_WORKING";
    break;
  case OTX_DEVICE_BLOCKED_AUTO:
    szDevState = "OTX_DEVICE_BLOCKED_AUTO"; 
    break;
  case OTX_DEVICE_BLOCKED_MANUAL:
    szDevState = "OTX_DEVICE_BLOCKED_MANUAL";
    break;
  default:
    szDevState = "Invalid Device state";
    break;
  }
  DemoLogf("Device State(0x%X): %s\n", hHandle, szDevState);
}
 



/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (result == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (result == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  OTX_UINT32 i;
  OTX_UINT8  buf[1024];
  OTX_UINT32 nFrameLen;
  OTX_UINT32 nFrameSeq;
  OtxHdlcFrameEndCodeE nFrameEndCode;
  OTX_DATETIME nTimeStamp;
  char msgStr[256];
  
  switch (pEventData->m_eDeviceType) {
  case OTX_LDEVICE_HDLC_RECEIVER:
    if (g_bHdlcVerbose) {
      DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    }
    break;
  default:
    DemoPrintTimeStamp(pEventData->m_nTimeStamp);
    break;
  }


  switch (pEventData->m_eDeviceType) {
  case OTX_DEVICE_TSS:
    // Event from the Time-Space Switch
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("TSS[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("TSS[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;
    
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DSP[%d] Event: %s (m_nParam=0x%X, m_nRequestId=0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DIALER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      DemoLogf("DtmfDialler[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    } else {
      DemoLogf("DtmfDialler[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_TONE_DTMF_DETECTOR:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from Dtmf Detector
      if (pEventData->m_nCode == OTX_DTMF_DETECTOR_EC_DIGIT) {
        DemoLogf("DtmfDetector[%d] detected DTMF digit %d \n", pEventData->m_nSourceId, pEventData->m_nParam);
      } else {
        // Unexpected event from this device
        DemoLogf("DtmfDetector[%d] Unexpected Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("DtmfDetector[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_RECEIVER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      // Event from HLDC Receiver
      // @todo DSP and LI HdlcReceivers should return the same event code since they are of the same logical device type
      if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {

        otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        while (otxRc == OTX_S_OK) {
          
          assert(pEventData->m_nSourceId < DEMO_LOGIC_DEVS_MAX);

          if (pEventData->m_nSourceId < DEMO_LOGIC_DEVS_MAX) {
            g_coHdlcRxFrames[pEventData->m_nSourceId]++;
          }

          
          if (g_bHdlcVerbose) {
            DemoLogf("HDLC Frame Received [Dsp=%d, Li=%d, Ts=%d] Sequence Number = %d\n", g_sqReceivers[pEventData->m_nSourceId].nDsp, g_sqReceivers[pEventData->m_nSourceId].nLi, g_sqReceivers[pEventData->m_nSourceId].nThorTs, pEventData->m_nParam);
            
            DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
            for (i = 0; i < nFrameLen && i < 1024; i++) {
              DemoLogf("%.2X ", buf[i]);
            }
            DemoLogf("\n");
          } else {
            if (nFrameEndCode) {
              DemoLogf("HdlcReceiver[%d] Len=%d, Fe=0x%.2X, Sq=%d: ", pEventData->m_nSourceId, nFrameLen, nFrameEndCode, nFrameSeq);
              for (i = 0; i < nFrameLen && i < 256; i++) {
                DemoLogf("%.2X ", buf[i]);
              }
              DemoLogf("\n");
            }
          }
          otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, 1024, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
        }
        if (otxRc != OTX_S_FALSE) {
          DemoLogf("HdlcReceiver[%d] OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
          DemoCheckRc(pEventData->m_hDevice, otxRc);
        }
      } else {
        // Unexpected event from this device
        DemoLogf("HdlcReceiver[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("HdlcReceiver[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}

/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentHdlcDev = 1;
  static OTX_INT32      nCurrentLi = 0;
  OTX_RESULT result;

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" 0-7 Select current Li\n");
    DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
#ifdef DEMO_SPM1_HDLC_SENDER
    DemoLogf(" t   Send an HDLC frame from current HDLC Sender\n");
#endif
    DemoLogf(" c   Set Clock Source\n");
    DemoLogf(" l   Query Line Status of all Lis\n");
    DemoLogf(" L   Set HDLC Frame length to be transmitted\n");
    DemoLogf(" s   Display the SS7 HDLC receive status of the current HDLC Receiver\n");
    DemoLogf(" e   Print the state of the opened devices\n");
    DemoLogf(" g   Get current time\n");
    break;

#if 1 // Only for debugging
  case 'o':
    {
      char szBuf[12];
      OTX_UINT32 nCmd;
      OTX_UINT8 nTs;
      OTX_UINT8 nHighway;
      OTX_UINT8 nSinkOrSource;
      
      DemoLogf("Amb Driver Debug Command. Which Cmd?\n");
      DemoLogf("0: Dump Local Connection Memory\n");
      DemoLogf("1: Dump all three CAMs\n");
      DemoLogf("2: Clear Connections in CAM-Os\n");
      DemoLogf("3: Find CAM entry in CAM-Os\n");
      OsaConsoleGetStr(szBuf);
      nCmd = (OTX_INT32) strtol(szBuf, NULL, 0);

      if (nCmd == 3) {
        DemoLogf("Enter nTs: ");
        OsaConsoleGetStr(szBuf);
        nTs = (OTX_UINT8) strtol(szBuf, NULL, 0);

        DemoLogf("Enter nHighway (0-47): ");
        OsaConsoleGetStr(szBuf);
        nHighway = (OTX_UINT8) strtol(szBuf, NULL, 0);

        nCmd |= (nHighway << 3*8) | (nTs << 2*8);

        DemoLogf("Enter Sink[1] or Source[0]: ");
        OsaConsoleGetStr(szBuf);
        nSinkOrSource = (OTX_UINT8) strtol(szBuf, NULL, 0);

        if (nSinkOrSource == 1) {
          nCmd |= 0x80000000;
        }

      }
      
      DemoLogf("Brd Thor: Issuing Driver Debug Command %X\n", nCmd);
      OtxDrvDbgDump(g_hPhysDevsThor.hTss, nCmd);
      DemoLogf("Brd Vidar: Issuing Driver Debug Command %X\n", nCmd);
      OtxDrvDbgDump(g_hPhysDevsVidar.hTss, nCmd);
      DemoLogf("Done.\n");
    }
    break;
#endif

  case 'g':
    {
      OTX_DATETIME nTimeStamp;
      OTX_DATETIME nTimeStampMs;
      OTX_DATETIME nTimeStampSec;
      struct tm *ptCurrentDayAndTime;
      time_t nCurrentTimeSecs;
      
      result = OtxDrvGetCurrentTime(&nTimeStamp); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

      //DemoLogf("%.8X%.8X (Current time)\n", nTimeStamp.sUint32.nHigh, nTimeStamp.sUint32.nLow);

      // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
			#if defined(__linux__)
				nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
			#elif defined(WIN32)
				nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
			#endif

      // Convert to seconds since Jan 1, 1970 (so we can use this time with localtime())
      nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;

      // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
      nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);

      nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;
      ptCurrentDayAndTime = localtime(&nCurrentTimeSecs);
      
      if (ptCurrentDayAndTime) {
        
        DemoLogf("Current time is (%.2d:%.2d:%.2d.%.3d) %s", 
          ptCurrentDayAndTime->tm_hour,
          ptCurrentDayAndTime->tm_min, 
          ptCurrentDayAndTime->tm_sec,
          nTimeStampMs.sUint32.nLow,
          asctime(ptCurrentDayAndTime)
          );
      }

    }

    break;

#if 0 // for driver debugging
  /*
  case 'y':
    {
      OTX_UINT32 readVal;
      OTX_UINT32 i;

      DemoLogf("Dumping registers in Li%d \n", nCurrentLi);

      for (i = 0; i < 115; i++) {
        result = OtxDrvReadReg8(g_hPhysDevsThor.hLi[nCurrentLi], i, &readVal); DemoCheckRc(g_hPhysDevsThor.hLi[nCurrentLi], result);
        DemoLogf("Reg%.3d=%X\n", i, (OTX_UINT8)readVal);
      }

      DemoLogf("Done.\n");
    }
    break;
  */
#endif


  case 'e':
    {
      OTX_INT32 i = 0;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("E1T1#%d ", i); DemoPrintDeviceState(g_hPhysDevsThor.hLi[i]); DemoLogf("\n");
      }

      for (i = 0; i < DEMO_DSP_MAX; i++) {
        DemoLogf("DSP#%d ", i); DemoPrintDeviceState(g_hPhysDevsVidar.hDsp[i]); DemoLogf("\n");
      }

      for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
        DemoLogf("HDLC Receiver#%d ", i); DemoPrintDeviceState(g_sqReceivers[i].hHandle); DemoLogf("\n");
      }
    }
    break;



  case 'v':
    {
      OTX_INT32 i;
      g_bHdlcVerbose = !g_bHdlcVerbose;
      DemoLogf("g_bHdlcVerbose is now %s.\n", g_bHdlcVerbose ? "TRUE": "FALSE");
      for (i = 0; i < DEMO_LOGIC_DEVS_MAX; i++) {
        if (g_coHdlcRxFrames[i]) {
          DemoLogf("g_coHdlcRxFrames[%d] = %d (reset to 0) \n", i, g_coHdlcRxFrames[i]);
          g_coHdlcRxFrames[i] = 0;
        }
      }
    }
    break;

  case 's':
    {
      OTX_UINT32  nFisuCount;   // Number of FISUs received
      OTX_UINT32  nLssuCount;   // Number of LSSUs received
      OTX_UINT32  nMsuCount;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.
      DemoLogf("SS7 Status for Spm1Receiver #%d \n", nCurrentHdlcDev);

      result = OtxHdlcSS7GetReceiveStatus(g_sqReceivers[nCurrentHdlcDev].hHandle, &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_sqReceivers[nCurrentHdlcDev].hHandle, result);

      if (OTX_SUCCESS(result)) {
        DemoLogf("FISUs received: %lu\n", nFisuCount);
        DemoLogf("LSSUs received: %lu\n", nLssuCount);
        DemoLogf("MSUs  received: %lu\n", nMsuCount);
        DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
      }

    }
    break;

  case 'L':
    {
      char strBuf[12];
      DemoLogf("Enter HDLC Frame Length to be transmitted (%d) ", g_nMsgLen);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        g_nMsgLen = (OTX_INT32)strtol(strBuf, NULL, 0);
        DemoLogf("HDLC Frame Length is now set to %d\n", g_nMsgLen);
      } else {
        DemoLogf("HDLC Frame Length remains unchanged\n");
      }
    }
    break;


  case 'l':
    {
      OTX_INT32 i = 0;
      OtxT1E1Layer1StatusE eLayer1Status;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevsThor.hLi[i], &eLayer1Status); DemoCheckRc(g_hPhysDevsThor.hLi[i], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

#if 1 // Only for debugging
  case 'd':
    {
      OTX_INT32 nCurrentDsp = 0;
      
      DemoLogf("Dump DSP#%d debug data\n", nCurrentDsp);
      OtxDrvDbgDump(g_hPhysDevsVidar.hDsp[nCurrentDsp], 0);
    }
    break;

  case 'b':
    {
      OTX_INT32 nCurrentDsp = 0;
      
      DemoLogf("Init Bsp on DSP#%d\n", nCurrentDsp);
      OtxDrvDbgDump(g_hPhysDevsVidar.hDsp[nCurrentDsp], 1);
    }
    break;
#endif

  case 'z':
    {
      OtxT1E1TermModeE eTermMode;

      DemoLogf("Set Line Termination mode:\n");
      DemoLogf("0 = OTX_T1E1_TERM_120_OHM\n");
      DemoLogf("1 = OTX_T1E1_TERM_100_OHM\n");
      DemoLogf("2 = OTX_T1E1_TERM_75_OHM\n");
      DemoLogf("3 = OTX_T1E1_TERM_HIGHZ\n");
      switch (getch() - '0') {
      case 0:
        eTermMode = OTX_T1E1_TERM_120_OHM;
        break;
      case 1:
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      case 2:
        eTermMode = OTX_T1E1_TERM_75_OHM;
        break;
      case 3:
        eTermMode = OTX_T1E1_TERM_HIGHZ;
        break;
      default:
        DemoLogf("Invalid entry. Setting Line Termination to OTX_T1E1_TERM_100_OHM\n");
        eTermMode = OTX_T1E1_TERM_100_OHM;
        break;
      }

      

      result = OtxT1E1SetLineTermination(g_hPhysDevsThor.hLi[nCurrentHdlcDev], eTermMode); DemoCheckRc(g_hPhysDevsThor.hLi[nCurrentHdlcDev], result);
    }
    break;


#ifdef DEMO_SPM1_HDLC_SENDER
  case 't':
    {
      OTX_UINT8 sqHdlcFrame[OD_HDLC_MAX_TX_HDLC_FRAME_LEN];
      OTX_INT32 i;
      OTX_HANDLE hHandle;

      hHandle = g_sqSenders[nCurrentHdlcDev].hHandle;

      for (i = 0; i < g_nMsgLen; i++) {
        sqHdlcFrame[i] = (OTX_UINT8) i;
      }

      DemoLogf("Sending HDLC frame from HDLC Sender #%d [Dsp=%d, Li=%d, Ts=%d]...", nCurrentHdlcDev, g_sqSenders[nCurrentHdlcDev].nDsp, g_sqSenders[nCurrentHdlcDev].nLi, g_sqSenders[nCurrentHdlcDev].nThorTs);

      // Send HDLC Frame
      result = OtxHdlcSendData(hHandle, sqHdlcFrame, g_nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
#endif


  case 'c':
    {
      OtxBrdClockSourceE  eClkSource;
      OtxBrdClockModeE    eClkMode;
      DemoLogf("Set Clock Source for Thor-2-PCI \n");
      DemoLogf("0 = OTX_CLOCK_SOURCE_INTERNAL  - LIs synchronized to on-board oscillator\n");
      DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_0   - LIs synchronized to FA0 receive span\n");
      DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_1   - LIs synchronized to FA1 receive span\n");
      DemoLogf("3 = OTX_CLOCK_SOURCE_A_CLOCKS  - H.100 Slave\n");

      switch (getch() - '0') {
      case 0:
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      case 1:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      case 2:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      case 3:
        eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      default:
        DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        eClkMode = OTX_CLOCK_MODE_LOCAL;
        break;
      }

      result = OtxBrdSetClocks(g_hPhysDevsThor.hThorController0, eClkMode, eClkSource); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);

      // Setting the clock source also performs a reset of the E1T1 devices. Therefore we need to reconfigure the E1T1 devices here.
      DemoConfigureE1T1Devices();

      DemoLogf("Done.\n");
    }
    break;
  
  case 'C':
    {
      OtxBrdClockSourceE  eClkSource;
      OtxBrdClockModeE    eClkMode;
      DemoLogf("Set Clock Source for Vidar-5x16-PCI \n");
      DemoLogf("0 = OTX_CLOCK_SOURCE_INTERNAL  - LIs synchronized to on-board oscillator\n");
      DemoLogf("1 = OTX_CLOCK_SOURCE_LOCAL_0   - LIs synchronized to FA0 receive span\n");
      DemoLogf("2 = OTX_CLOCK_SOURCE_LOCAL_1   - LIs synchronized to FA1 receive span\n");
      DemoLogf("3 = OTX_CLOCK_SOURCE_A_CLOCKS  - H.100 Slave\n");

      switch (getch() - '0') {
      case 0:
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      case 1:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      case 2:
        eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      case 3:
        eClkSource = OTX_CLOCK_SOURCE_A_CLOCKS;
        eClkMode = OTX_CLOCK_MODE_H100;
        break;
      default:
        DemoLogf("Invalid entry. Setting Clock Source to OTX_CLOCK_SOURCE_INTERNAL\n");
        eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
        eClkMode = OTX_CLOCK_MODE_LOCAL;
        break;
      }

      result = OtxBrdSetClocks(g_hPhysDevsVidar.hVidarController0, eClkMode, eClkSource); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);

      DemoLogf("Done.\n");
    }
    break;
  
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
    {
      // Select current Li
      OTX_INT32 nLi = commandChar - '0';

      if (nLi >= g_demoEnv.coLi) {
        DemoLogf("ERROR: Li#%d is not supported on this board\n", nLi);
      } else {
        nCurrentLi = nLi;
        DemoLogf("Li#%d is selected\n", nLi);
      }
    }
    break;

  case 'n':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev++;
      if (nCurrentHdlcDev >= DEMO_LOGIC_DEVS_MAX) {
        // Wrap 
        nCurrentHdlcDev = 0;
      }
      
      DemoLogf("DSP HDLC Sender/Receiver#%d [Dsp=%d, Li=%d, Ts=%d] Selected \n", nCurrentHdlcDev, g_sqSenders[nCurrentHdlcDev].nDsp, g_sqSenders[nCurrentHdlcDev].nLi, g_sqSenders[nCurrentHdlcDev].nThorTs);
      
    }
    break;
        
  case 'N':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev--;
      if (nCurrentHdlcDev < 0) {
        // Wrap
        nCurrentHdlcDev = DEMO_LOGIC_DEVS_MAX-1;
      }
      
      DemoLogf("DSP HDLC Sender/Receiver#%d [Dsp=%d, Li=%d, Ts=%d] Selected \n", nCurrentHdlcDev, g_sqSenders[nCurrentHdlcDev].nDsp, g_sqSenders[nCurrentHdlcDev].nLi, g_sqSenders[nCurrentHdlcDev].nThorTs);
    }
    break;
           

  case 'q':
  case 'Q':
    return 1;
    break;


  case 'r':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_INT32 nCurrentDsp = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentDsp);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevsThor.hLi[nCurrentDsp], nAddress, &readVal); DemoCheckRc(g_hPhysDevsThor.hLi[nCurrentDsp], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentDsp, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 'R':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevsThor.hThorController0, nAddress, &readVal); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;

  case 'G':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;

      DemoLogf("Read from a 16-bit register in the Vidar-5x16-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg16(g_hPhysDevsVidar.hVidarController0, nAddress, &readVal); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);

      DemoLogf("Brd Address 0x%X = %.4X\n", nAddress, (OTX_UINT16)readVal);
    }
    break;

  case 'w':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_INT32 nCurrentDsp = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentDsp);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevsThor.hLi[nCurrentDsp], nAddress, nValue); DemoCheckRc(g_hPhysDevsThor.hLi[nCurrentDsp], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentDsp, nAddress);
    }
    break;

  case 'W':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the Thor-2-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevsThor.hThorController0, nAddress, nValue); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;

  case 'S':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;

      DemoLogf("Write to a 16-bit register in the Vidar-5x16-PCI device.\n");
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg16(g_hPhysDevsVidar.hVidarController0, nAddress, nValue); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);

      DemoLogf("Wrote %.4X to Brd Address 0x%X\n", (OTX_UINT16)nValue, nAddress);
    }
    break;


  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->bSs7FisuFiltering = OTX_FALSE;
  envP->nBoardType        = OTX_DEVICE_THOR_PCI;
  envP->coLi              = 2;
  
  for (i = 1; i < argc; i++) {        // Scan command line arguments
    OsaStdStrUpr(argv[i]);
    if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
      if (*(argv[i]+1) == 'L') {   // Log file
        envP->logFileName = argv[i]+2;
      } else if (*(argv[i]+1) == 'B') {   // Board number
        envP->boardNo = strtoul(argv[i]+2, NULL, 0);
      } else if (*(argv[i]+1) == 'T') {   // Board Type
        if ( *(argv[i]+2) == '2') {
          envP->nBoardType = OTX_DEVICE_THOR_PCI;
          envP->coLi = 2;
        } else if ( *(argv[i]+2) == '8') {
          envP->nBoardType  = OTX_DEVICE_THOR_8_PCI;
          envP->coLi = 8;
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          exit(1);
        }
      } else if (*(argv[i]+1) == 'S') {   // SS7 FISU filtering
        envP->bSs7FisuFiltering = OTX_TRUE;
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
				exit(1);
      }
    } else {
      DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
      DemoHelpPanel();
      OsaConsoleExit();
			exit(1);
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>  : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /T[2|8]      : Board Type used for the demo, where\n");
  DemoLogf("                  2 = Use Thor-2-PCI Board\n");
  DemoLogf("                  8 = Use Thor-8-PCI Board\n");
  DemoLogf(" /L<fileName> : Log file name\n");
  DemoLogf(" /S           : SS7 FISU filtering active\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;
  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

}



/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

	OsaConsoleInit();

  DemoLogf("T2p5x16, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("HDLC pipes over H.100 Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);

  DemoInitalizeGlobalData();

  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();
  DemoLogf("Enabled physical devices...\n");

  // Set Li0 on Thor-2-PCI as the H.100 Clock Master
  result = OtxBrdSetClocks(g_hPhysDevsThor.hThorController0,   OTX_CLOCK_MODE_H100, OTX_CLOCK_SOURCE_LOCAL_0); DemoCheckRc(g_hPhysDevsThor.hThorController0, result);
  OsaThreadSleep(1000);
  // Set Vidar-5x16-PCI as an H.100 Clock Slave (synchronize to the H.100 A-clock sourced by the Thor-2-PCI board)
  result = OtxBrdSetClocks(g_hPhysDevsVidar.hVidarController0, OTX_CLOCK_MODE_H100,  OTX_CLOCK_SOURCE_A_CLOCKS); DemoCheckRc(g_hPhysDevsVidar.hVidarController0, result);
  OsaThreadSleep(1000);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing Logical devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Main Loop. Loop here and wait for Events from the driver or commands from the user ....
	OsaThreadStart(DemoIdleFunction);

  DemoLogf("\nPress 'h' for help\n");

  while (!exitFlag) {
    commandChar = getch();
    exitFlag = DemoProcessUserCommand(commandChar);
  }

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}



